import morepath

from collections import defaultdict


def keywords_encode(d):
    """ Takes a dictionary of keywords and encodes them into a somewhat
    readable url query format.

    For example:

        {
            'color': ['blue', 'red'],
            'weight': ['normal']
        }

    Results in

        '+color:blue+color:red+weight:normal'

    Instead of a dictionary we can also use any kind of object
    which has a 'keywords' property returning the expected dictionary.

    Note that that object won't be recreated during decode however.

    """

    if not d:
        return ''

    if hasattr(d, 'keywords'):
        d = d.keywords

    return '+'.join('{}:{}'.format(k, v) for k in d for v in d[k])


def keywords_decode(s):
    """ Deocdes keywords creaged by :func:`keywords_encode`. """

    if not s:
        return None

    result = defaultdict(list)

    for item in s.split('+'):
        key, value = item.split(':', 1)
        result[key].append(value)

    return result


keywords_converter = morepath.Converter(
    decode=keywords_decode, encode=keywords_encode
)



from collections import namedtuple
from onegov.chat import Message
from onegov.chat import MessageCollection
from onegov.core.custom import json
from onegov.core.security import Private
from onegov.org import OrgApp, _
from onegov.org.layout import MessageCollectionLayout
from onegov.user import User
from sqlalchemy import inspect


class Owner(namedtuple('OwnerBase', ('username', 'realname'))):

    @property
    def initials(self):
        return User.get_initials(self.username, self.realname)

    @property
    def name(self):
        return self.realname or self.username


@OrgApp.json(model=MessageCollection, permission=Private, name='feed')
def view_messages_feed(self, request):
    mapper = inspect(Message).polymorphic_map
    layout = MessageCollectionLayout(self, request)

    def cast(message):
        message.__class__ = mapper[message.type].class_
        return message

    messages = tuple(cast(m) for m in self.query())
    usernames = {m.owner for m in messages if m.owner}

    if usernames:
        q = request.session.query(User)
        q = q.with_entities(User.username, User.realname)
        q = q.filter(User.username.in_(usernames))

        owners = {u.username: Owner(u.username, u.realname) for u in q}
        owners.update({
            username: Owner(username, None)
            for username in usernames
            if username not in owners
        })
    else:
        owners = {}

    return {
        'messages': [
            {
                'id': m.id,
                'type': m.type,
                'date': ', '.join((
                    layout.format_date(m.created, 'weekday_long'),
                    layout.format_date(m.created, 'date_long')
                )),
                'html': m.get(
                    request,
                    owner=owners.get(m.owner),
                    layout=layout
                ),
            } for m in messages
        ]
    }


@OrgApp.html(
    model=MessageCollection,
    permission=Private,
    template='timeline.pt'
)
def view_messages(self, request):

    # The initial load contains only the 25 latest message (the feed will
    # return the 25 oldest messages by default)
    if not self.newer_than:
        beyond_horizon = self.latest_message(offset=25)

        if beyond_horizon:
            self.newer_than = beyond_horizon.id

    return {
        'layout': MessageCollectionLayout(self, request),
        'title': _("Timeline"),
        'feed': request.link(self, 'feed'),
        'feed_data': json.dumps(
            view_messages_feed(self, request)
        ),
        'feed_interval': 15
    }



from cached_property import cached_property
from onegov.form import Form, FormDefinition
from onegov.org import _
from wtforms.fields import BooleanField, RadioField
from wtforms.fields.html5 import DateField, IntegerField
from wtforms.validators import NumberRange, InputRequired, ValidationError


class FormRegistrationWindowForm(Form):
    """ Form to edit registration windows. """

    start = DateField(
        label=_("Start"),
        validators=[InputRequired()]
    )

    end = DateField(
        label=_("End"),
        validators=[InputRequired()]
    )

    limit_attendees = RadioField(
        label=_("Limit the number of attendees"),
        fieldset=_("Attendees"),
        choices=[
            ('yes', _("Yes, limit the number of attendees")),
            ('no', _("No, allow an unlimited number of attendees"))
        ],
        default='yes'
    )

    limit = IntegerField(
        label=_("Number of attendees"),
        fieldset=_("Attendees"),
        depends_on=('limit_attendees', 'yes'),
        validators=(NumberRange(min=1, max=None), )
    )

    waitinglist = RadioField(
        label=_("Waitinglist"),
        fieldset=_("Attendees"),
        depends_on=('limit_attendees', 'yes'),
        choices=[
            ('yes', _("Yes, allow for more submissions than available spots")),
            ('no', _("No, ensure that all submissions can be confirmed"))
        ],
        default='yes'
    )

    stop = BooleanField(
        label=_("Do not accept any submissions"),
        fieldset=_("Advanced"),
        default=False
    )

    def process_obj(self, obj):
        super().process_obj(obj)
        self.waitinglist.data = obj.overflow and 'yes' or 'no'
        self.limit_attendees.data = obj.limit and 'yes' or 'no'
        self.limit.data = obj.limit or ''
        self.stop.data = not obj.enabled

    def populate_obj(self, obj, *args, **kwargs):
        super().populate_obj(obj, *args, **kwargs)
        obj.overflow = self.waitinglist.data == 'yes'
        obj.limit = self.limit_attendees.data == 'yes' and self.limit.data or 0
        obj.enabled = not self.stop.data

    @cached_property
    def claimed_spots(self):
        return self.model.claimed_spots

    @cached_property
    def requested_spots(self):
        return self.model.requested_spots

    def validate_limit(self, field):

        # new registration windows do not need to enforce limits
        if isinstance(self.model, FormDefinition):
            return

        # nor do unlimited registration windows
        if not self.limit_attendees.data:
            return

        if self.limit.data < self.claimed_spots:
            raise ValidationError(_(
                "The limit cannot be lower than the already confirmed "
                "number of attendees (${claimed_spots})",
                mapping={
                    'claimed_spots': self.claimed_spots
                }
            ))

        # without a waitinglist we guarantee that all people who sent in
        # an attendance request are getting a spot (unless they are maunally
        # cancelled), so we need to take requested spots into account
        if self.waitinglist.data == 'yes':
            return

        if self.limit.data < (self.requested_spots + self.claimed_spots):
            raise ValidationError(_(
                "The limit cannot be lower than the already confirmed "
                "number attendees (${claimed_spots}) and the number of "
                "pending requests (${pending_requests}). Either enable the "
                "waiting list, process the pending requests or increase the "
                "limit. ",
                mapping={
                    'claimed_spots': self.claimed_spots,
                    'pending_requests': self.requested_spots
                }
            ))



from collections import defaultdict
from onegov.core.security import Private
from onegov.form import FormDefinition
from onegov.form import FormRegistrationWindow
from onegov.form import FormSubmission
from onegov.org import OrgApp, _
from onegov.org.forms import FormRegistrationWindowForm
from onegov.org.layout import FormSubmissionLayout
from onegov.org.new_elements import Link, Confirm, Intercooler
from sqlalchemy import desc


@OrgApp.form(
    model=FormDefinition,
    name='new-registration-window',
    permission=Private,
    form=FormRegistrationWindowForm,
    template='form.pt')
def handle_new_registration_form(self, request, form):

    title = _("New Registration Window")

    layout = FormSubmissionLayout(self, request)
    layout.editbar_links = None
    layout.breadcrumbs.append(Link(title, '#'))

    if form.submitted(request):

        form.populate_obj(self.add_registration_window(
            form.start.data,
            form.end.data
        ))

        request.success(_("The registration window was added successfully"))
        return request.redirect(request.link(self))

    return {
        'layout': layout,
        'title': title,
        'form': form,
        'helptext': _(
            "Registration windows limit forms to a set number of submissions "
            "and a specific time-range."
        )
    }


@OrgApp.html(
    model=FormRegistrationWindow,
    permission=Private,
    template='registration_window.pt')
def view_registration_window(self, request):

    layout = FormSubmissionLayout(self.form, request)
    title = layout.format_date_range(self.start, self.end)

    layout.breadcrumbs.append(Link(title, '#'))
    layout.editbar_links = [
        Link(
            text=_("Edit"),
            url=request.link(self, 'edit'),
            attrs={'class': 'edit-link'}
        ),
        Link(
            text=_("Delete"),
            url=layout.csrf_protected_url(request.link(self)),
            attrs={'class': 'delete-link'},
            traits=(
                Confirm(
                    _(
                        "Do you really want to delete "
                        "this registration window?"
                    ),
                    _("Existing submissions will be disassociated."),
                    _("Delete registration window")
                ),
                Intercooler(
                    request_method='DELETE',
                    redirect_after=request.link(self.form)
                )
            )
        )
    ]

    registrations = defaultdict(list)

    q = request.session.query(FormSubmission)
    q = q.filter_by(registration_window_id=self.id)
    q = q.filter_by(state='complete')
    q = q.order_by(desc(FormSubmission.received))

    for submission in q:
        if not submission.registration_state:
            continue

        registrations[submission.registration_state].append(submission)

    return {
        'layout': layout,
        'title': title,
        'model': self,
        'registrations': registrations,
        'groups': (
            (_("Open"), 'open'),
            (_("Confirmed"), 'confirmed'),
            (_("Cancelled"), 'cancelled'),
        )
    }


@OrgApp.form(
    model=FormRegistrationWindow,
    permission=Private,
    form=FormRegistrationWindowForm,
    template='form.pt',
    name='edit')
def handle_edit_registration_form(self, request, form):

    title = _("Edit Registration Window")

    layout = FormSubmissionLayout(self.form, request)
    layout.breadcrumbs.append(Link(title, '#'))
    layout.editbar_links = []

    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return request.redirect(request.link(self.form))

    elif not request.POST:
        form.process(obj=self)

    return {
        'layout': layout,
        'title': title,
        'form': form
    }


@OrgApp.view(
    model=FormRegistrationWindow,
    permission=Private,
    request_method='DELETE')
def delete_registration_window(self, request):
    request.assert_valid_csrf_token()

    self.disassociate()
    request.session.delete(self)

    request.success(_("The registration window was deleted"))



from collections import defaultdict
from copy import copy
from onegov.core.crypto import random_password
from onegov.core.directives import query_form_class
from onegov.core.security import Secret
from onegov.core.templates import render_template
from onegov.form import merge_forms
from onegov.org import _, OrgApp
from onegov.org.forms import ManageUserForm, NewUserForm
from onegov.org.layout import DefaultMailLayout
from onegov.org.layout import UserLayout
from onegov.org.layout import UserManagementLayout
from onegov.org.new_elements import Link, LinkGroup
from onegov.ticket import TicketCollection, Ticket
from onegov.user import Auth, User, UserCollection
from onegov.user.errors import ExistingUserError
from onegov.user.forms import SignupLinkForm
from wtforms.validators import Optional


@OrgApp.html(model=UserCollection, template='usermanagement.pt',
             permission=Secret)
def view_usermanagement(self, request):
    """ Allows the management of organisation users. """

    layout = UserManagementLayout(self, request)

    users = defaultdict(list)

    for user in self.query().order_by(User.username).all():
        users[user.role].append(user)

    filters = {}

    filters['role'] = [
        Link(
            text=request.translate(title),
            active=value in self.filters.get('role', tuple()),
            url=request.link(self.for_filter(role=value))
        ) for title, value in (
            (_("Administrator"), 'admin'),
            (_("Editor"), 'editor'),
            (_("Member"), 'member'),
        )
    ]

    filters['active'] = [
        Link(
            text=request.translate(title),
            active=value in self.filters.get('active', tuple()),
            url=request.link(self.for_filter(active=value))
        ) for title, value in (
            (_("Active"), True),
            (_("Inactive"), False)
        )
    ]

    filters['tag'] = [
        Link(
            text=tag,
            active=tag in self.filters.get('tag', tuple()),
            url=request.link(self.for_filter(tag=tag))
        ) for tag in self.tags
    ]

    return {
        'layout': layout,
        'title': _("User Management"),
        'users': users,
        'filters': filters
    }


@OrgApp.form(
    model=UserCollection,
    template='signup_link.pt',
    permission=Secret,
    form=SignupLinkForm,
    name='signup-link')
def handle_create_signup_link(self, request, form):
    link = None

    if form.submitted(request):
        auth = Auth.from_app(request.app)
        auth.signup_token = form.signup_token(auth)

        link = request.link(auth, 'register')

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(_("New Signup Link"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("New Signup Link"),
        'link': link,
        'form': form
    }


@OrgApp.html(model=User, template='user.pt', permission=Secret)
def view_user(self, request):
    """ Shows all objects owned by the given user. """

    layout = UserLayout(self, request)

    linkgroups = [
        fn(request, self) for fn in request.app.config.linkgroup_registry
    ]
    linkgroups.sort(key=lambda group: request.translate(group.title))

    return {
        'layout': layout,
        'title': self.title,
        'linkgroups': linkgroups
    }


@OrgApp.userlinks()
def ticket_links(request, user):
    tickets = TicketCollection(request.session).query()
    tickets = tickets.filter_by(user_id=user.id)
    tickets = tickets.order_by(Ticket.number)
    tickets = tickets.with_entities(
        Ticket.id, Ticket.number, Ticket.handler_code)

    return LinkGroup(
        title=_("Tickets"),
        links=[
            Link(
                ticket.number,
                request.class_link(Ticket, {
                    'handler_code': ticket.handler_code,
                    'id': ticket.id
                }),
            )
            for ticket in tickets
        ]
    )


def get_manage_user_form(self, request):
    userprofile_form = query_form_class(request, self, name='userprofile')
    assert userprofile_form

    class OptionalUserprofile(userprofile_form):

        hooked = False

        def submitted(self, request):
            # fields only present on the userprofile_form are made optional
            # to make sure that we can always change the active/inactive state
            # of the user and the role the user has
            if not self.hooked:
                for name, field in self._fields.items():
                    if not hasattr(userprofile_form, name):
                        continue

                    if not field.validators:
                        continue

                    # be careful not to change the class itself
                    field.validators = copy(field.validators)
                    field.validators.insert(0, Optional())

                self.hooked = True

            return super().submitted(request)

    return merge_forms(ManageUserForm, OptionalUserprofile)


@OrgApp.form(model=User, template='form.pt', form=get_manage_user_form,
             permission=Secret, name='edit')
def handle_manage_user(self, request, form):

    # XXX the manage user form doesn't have access to the username
    # because it can't be edited, so we need to inject it here
    # for validation purposes (check for a unique yubikey)
    form.current_username = self.username

    if not request.app.enable_yubikey:
        form.delete_field('yubikey')

    if form.submitted(request):
        form.populate_obj(self)
        request.success(_("Your changes were saved"))

        return request.redirect(request.class_link(UserCollection))

    elif not request.POST:
        form.process(obj=self)

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(self.username, '#'))

    return {
        'layout': layout,
        'title': self.username,
        'form': form
    }


@OrgApp.form(model=UserCollection, template='newuser.pt', form=NewUserForm,
             name='new', permission=Secret)
def handle_new_user(self, request, form):

    if not request.app.enable_yubikey:
        form.delete_field('yubikey')

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(_("New User"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        password = random_password()

        if form.data.get('yubikey'):
            second_factor = {
                'type': 'yubikey',
                'data': form.data['yubikey'][:12]
            }
        else:
            second_factor = None

        try:
            user = self.add(
                username=form.username.data,
                password=password,
                role=form.role.data,
                active=form.active,
                second_factor=second_factor
            )
        except ExistingUserError:
            form.username.errors.append(
                _("A user with this e-mail address already exists"))
        else:
            if form.send_activation_email.data:
                subject = request.translate(
                    _("An account was created for you")
                )

                content = render_template('mail_new_user.pt', request, {
                    'user': user,
                    'org': request.app.org,
                    'layout': DefaultMailLayout(user, request),
                    'title': subject
                })

                request.app.send_transactional_email(
                    subject=subject,
                    receivers=(user.username, ),
                    content=content,
                )

            request.info(_("The user was created successfully"))

            return {
                'layout': layout,
                'title': _("New User"),
                'username': form.username.data,
                'password': password,
                'sent_email': form.send_activation_email.data
            }

    return {
        'layout': layout,
        'title': _("New User"),
        'form': form,
        'password': None,
        'sent_email': False
    }
